<?php
namespace Ep\TecDoc;

use Ep\TecDoc\Record\ManufacturersRecord;
use Ep\TecDoc\Record\ModelSeriesRecord;
use Ep\TecDoc\Record\VehicleByIdRecord;
use Ep\TecDoc\Record\VehicleByIds4Record;
use Ep\TecDoc\Record\VehicleIdsByCriteriaRecord;
use Ep\TecDoc\Request\ManufacturersRequest;
use Ep\TecDoc\Request\ModelSeriesRequest;
use Ep\TecDoc\Request\RequestInterface;
use Ep\TecDoc\Request\VehicleByIds4Request;
use Ep\TecDoc\Request\VehicleIdsByCriteriaRequest;
use Ep\TecDoc\Response\ManufacturersResponse;
use Ep\TecDoc\Response\ModelSeriesResponse;
use Ep\TecDoc\Response\ResponseInterface;
use Ep\TecDoc\Response\VehicleByIds4Response;
use Ep\TecDoc\Response\VehicleIdsByCriteriaResponse;
use GuzzleHttp\ClientInterface as HttpClientInterface;
use GuzzleHttp\Exception\GuzzleException;
use Psr\Cache\CacheItemPoolInterface;
use Zend\Hydrator\HydratorInterface;
use Zend\Hydrator\HydratorPluginManager;
use Zend\Session\Container as SessionContainer;

/**
 * Class TecDoc
 *
 * @package Ep\TecDoc
 * @author Elektronika Perhoč <web@ep.hr>
 * @link http://gitlab.ep.local/ep-framework/tecdoc
 */
class TecDoc implements TecDocInterface
{
    const SERVICE_BASE_URI = 'https://webservice.tecalliance.services/pegasus-3-0/services/';
    const SERVICE_ENDPOINT = 'TecdocToCatDLB.jsonEndpoint';

    /** @var array */
    protected $config;
    /** @var HttpClientInterface */
    protected $http;
    /** @var CacheItemPoolInterface */
    protected $cache;
    /** @var HydratorPluginManager */
    protected $hydrators;

    /** @var VehicleByIdRecord */
    protected $currentVehicle;
    /** @var ManufacturersRecord[] */
    protected $featuredManufacturers;
    /** @var ManufacturersRecord[] */
    protected $manufacturers;
    /** @var ModelSeriesRecord[] */
    protected $models = [];
    /** @var VehicleIdsByCriteriaRecord[][] */
    protected $vehicleIdsByCriteria = [];
    /** @var SessionContainer */
    protected $session;

    /**
     * TecDoc constructor.
     *
     * @param array $config
     * @param HttpClientInterface $http
     * @param CacheItemPoolInterface $cache
     * @param HydratorPluginManager $hydrators
     * @param SessionContainer $session
     */
    public function __construct(
        array $config,
        HttpClientInterface $http,
        CacheItemPoolInterface $cache,
        HydratorPluginManager $hydrators,
        SessionContainer $session
    ) {
        $this->config = $config;
        $this->http = $http;
        $this->cache = $cache;
        $this->hydrators = $hydrators;
        $this->session = $session;
    }

    public function getConfig(): array
    {
        return $this->config;
    }

    public function setConfig(array $config): void
    {
        $this->config = $config;
    }

    /**
     * @return VehicleByIdRecord|null
     */
    public function getCurrentVehicle(): ?VehicleByIdRecord
    {
        /** @noinspection PhpUndefinedFieldInspection */
        return $this->currentVehicle;
    }

    /**
     * @param VehicleByIdRecord|int|string|null $vehicle
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function setCurrentVehicle($vehicle): void
    {
        if (is_string($vehicle) && preg_match('/(\d+)$/', $vehicle, $matches)) {
            $vehicle = intval($matches[1]);
        }

        if (is_integer($vehicle)) {
            $vehicle = $this->getVehicleById($vehicle);
        }

        /** @noinspection PhpUndefinedFieldInspection */
        $this->currentVehicle = $vehicle;
    }

    /**
     * @return ManufacturersRecord[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getFeaturedManufacturers(): array
    {
        if (null === $this->featuredManufacturers) {
            $cacheKey = 'manufacturers.featured';
            $cacheItem = $this->cache->getItem($cacheKey);

            if (!$cacheItem->isHit()) {
                $featured = [];

                if (count($this->config['featured_manufacturers']) > 0) {
                    $all = $this->getManufacturers();

                    foreach ($this->config['featured_manufacturers'] as $manufacturerId) {
                        $featured[$manufacturerId] = $all[$manufacturerId];
                    }
                }

                $cacheItem->set($featured);
                $this->cache->save($cacheItem);
            }

            $this->featuredManufacturers = $cacheItem->get();
        }

        return $this->featuredManufacturers;
    }

    /**
     * @return ManufacturersRecord[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getManufacturers(): array
    {
        if (null === $this->manufacturers) {
            $cacheKey = 'manufacturers.all';
            $cacheItem = $this->cache->getItem($cacheKey);

            if (!$cacheItem->isHit()) {
                $request = new ManufacturersRequest();
                /** @var ManufacturersResponse $response */
                $response = $this->request($request);
                $manufacturers = [];

                foreach ($response->getData() as $manufacturer) {
                    if (in_array($manufacturer->getId(), $this->config['manufacturers'])) {
                        $manufacturers[$manufacturer->getId()] = $manufacturer;
                    }
                }

                $cacheItem->set($manufacturers);
                $this->cache->save($cacheItem);
            }

            $this->manufacturers = $cacheItem->get();
        }

        return $this->manufacturers;
    }

    /**
     * @param ManufacturersRecord|int $manufacturer
     *
     * @return ModelSeriesRecord[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getModelSeries($manufacturer): array
    {
        if ($manufacturer instanceof ManufacturersRecord) {
            $manufacturer = $manufacturer->getId();
        } else {
            $manufacturer = intval($manufacturer);
        }

        if (!isset($this->models[$manufacturer])) {
            $cacheKey = 'models.manufacturer.' . $manufacturer;
            $cacheItem = $this->cache->getItem($cacheKey);

            if (!$cacheItem->isHit()) {
                $request = new ModelSeriesRequest();
                $request->setManuId($manufacturer);
                /** @var ModelSeriesResponse $response */
                $response = $this->request($request);
                $models = [];

                foreach ($response->getData() as $model) {
                    $models[] = $model;
                }

                $cacheItem->set($models);
                $this->cache->save($cacheItem);
            }

            $this->models[$manufacturer] = $cacheItem->get();
        }

        return $this->models[$manufacturer];
    }

    /**
     * @param ManufacturersRecord|int $manufacturer
     * @param ModelSeriesRecord|int $model
     *
     * @return VehicleIdsByCriteriaRecord[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getVehicleIdsByCriteria($manufacturer, $model): array
    {
        if ($manufacturer instanceof ManufacturersRecord) {
            $manufacturer = $manufacturer->getId();
        } else {
            $manufacturer = intval($manufacturer);
        }

        if ($model instanceof ModelSeriesRecord) {
            $model = $model->getId();
        } else {
            $model = intval($model);
        }

        if (!isset($this->vehicleIdsByCriteria[$manufacturer][$model])) {
            $cacheKey = sprintf('vehicleids.%d.%d', $manufacturer, $model);
            $cacheItem = $this->cache->getItem($cacheKey);

            if (!$cacheItem->isHit()) {
                if (!isset($this->vehicleIdsByCriteria[$manufacturer])) {
                    $this->vehicleIdsByCriteria[$manufacturer] = [];
                }

                $request = new VehicleIdsByCriteriaRequest();
                $request->setManuId($manufacturer);
                $request->setModId($model);
                $request->setCarType($this->config['linking_target_type']);
                $request->setCountriesCarSelection($this->config['country']);

                /** @var VehicleIdsByCriteriaResponse $response */
                $response = $this->request($request);
                $vehicles = [];

                foreach ($response->getData() as $vehicle) {
                    $vehicles[] = $vehicle;
                }

                $cacheItem->set($vehicles);
                $this->cache->save($cacheItem);
            }

            $this->vehicleIdsByCriteria[$manufacturer][$model] = $cacheItem->get();
        }

        return $this->vehicleIdsByCriteria[$manufacturer][$model];
    }

    /**
     * @param int $id
     *
     * @return VehicleByIdRecord
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getVehicleById(int $id): VehicleByIdRecord
    {
        $cacheKey = sprintf('vehicles.id.%d', $id);
        $cacheItem = $this->cache->getItem($cacheKey);

        if (!$cacheItem->isHit()) {
            $vehicles = $this->getVehicleByIds4([$id]);

            $cacheItem->set($vehicles[0]->getVehicleDetails());
            $this->cache->save($cacheItem);
        }

        return $cacheItem->get();
    }

    /**
     * @param int[] $vehicleIds
     *
     * @return VehicleByIds4Record[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getVehicleByIds4(array $vehicleIds): array
    {
        sort($vehicleIds);

        $cacheKey = sprintf('vehicles.byids4.%s', md5(serialize($vehicleIds)));
        $cacheItem = $this->cache->getItem($cacheKey);

        if (!$cacheItem->isHit()) {
            $chunked = array_chunk($vehicleIds, 25);
            $vehicles = [];

            foreach ($chunked as $chunk) {
                $request = new VehicleByIds4Request();
                $request->setCarIds($chunk);
                $request->setArticleCountry($this->config['country']);
                $request->setCountriesCarSelection($this->config['country']);

                /** @var VehicleByIds4Response $response */
                $response = $this->request($request);

                foreach ($response->getData() as $vehicle) {
                    $vehicles[] = $vehicle;
                }
            }

            $cacheItem->set($vehicles);
            $this->cache->save($cacheItem);
        }

        return $cacheItem->get();
    }

    /**
     * @param ManufacturersRecord|int $manufacturer
     * @param ModelSeriesRecord|int $model
     *
     * @return VehicleByIdRecord[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getVehicles($manufacturer, $model): array
    {
        if ($manufacturer instanceof ManufacturersRecord) {
            $manufacturer = $manufacturer->getId();
        } else {
            $manufacturer = intval($manufacturer);
        }

        if ($model instanceof ModelSeriesRecord) {
            $model = $model->getId();
        } else {
            $model = intval($model);
        }

        $cacheKey = sprintf('vehicles.all.%d.%d', $manufacturer, $model);
        $cacheItem = $this->cache->getItem($cacheKey);

        if (!$cacheItem->isHit()) {
            $vehicles = $this->getVehicleIdsByCriteria($manufacturer, $model);
            $vehicleIds = [];

            foreach ($vehicles as $vehicle) {
                $vehicleIds[] = $vehicle->getId();
            }

            $raw = $this->getVehicleByIds4($vehicleIds);
            $vehicles = [];

            foreach ($raw as $vehicle) {
                $vehicles[] = $vehicle->getVehicleDetails();
            }

            $cacheItem->set($vehicles);
            $this->cache->save($cacheItem);
        }

        return $cacheItem->get();
    }

    /**
     * @param RequestInterface $request
     *
     * @return ResponseInterface
     */
    protected function request(RequestInterface $request): ResponseInterface
    {
        $this->checkRequest($request);

        try {
            $response = $this->http->request('POST', self::SERVICE_ENDPOINT, [
                'json' => [
                    $request->getRemoteFunctionName() => $request->toArray(),
                ]
            ]);
        } catch (GuzzleException $e) {
            throw new Exception\HttpException(sprintf(
                'Došlo je do pogreške prilikom komunikacije s TecDoc-om: %s.',
                $e->getMessage()
            ));
        }

        $body = $response->getBody()->getContents();
        $data = json_decode($body, true);

        if (null === $data) {
            throw new Exception\JsonException(sprintf(
                'Došlo je do pogreške prilikom dekodiranja JSON-a: %s.',
                json_last_error_msg()
            ));
        }

        if (400 === $data['status']) {
            throw new Exception\BadRequestException(sprintf(
                'Request prema TecDoc-u je neispravan: %s.',
                $data['statusText']
            ));
        }

        $responseClass = $request->getResponseClass();

        if (!$this->hydrators->has($responseClass)) {
            throw new \RuntimeException(sprintf(
                'Ne mogu stvoriti hidrator za klasu %s',
                $responseClass
            ));
        }

        /** @var HydratorInterface $hydrator */
        $hydrator = $this->hydrators->get($responseClass);
        /** @var ResponseInterface $hydrated */
        $hydrated = $hydrator->hydrate($data, new $responseClass());

        return $hydrated;
    }

    /**
     * Dodaje nekakve parametre u request ako već nisu postavljeni.
     *
     * @param RequestInterface $request
     */
    protected function checkRequest(RequestInterface $request): void
    {
        if (!$request->getProvider()) {
            $request->setProvider($this->config['provider'] ?? 10);
        }

        if (
            method_exists($request, 'getLang') &&
            method_exists($request, 'setLang') &&
            !$request->getLang()
        ) {
            $request->setLang($this->config['lang'] ?? 'en');
        }

        if (
            method_exists($request, 'getCountry') &&
            method_exists($request, 'setCountry') &&
            !$request->getCountry()
        ) {
            $request->setCountry($this->config['country'] ?? 'HR');
        }

        if (
            method_exists($request, 'getLinkingTargetType') &&
            method_exists($request, 'setLinkingTargetType') &&
            !$request->getLinkingTargetType()
        ) {
            $request->setLinkingTargetType($this->config['linking_target_type'] ?? 'P');
        }
    }
}